import copy
import logging
import json
from http.server import BaseHTTPRequestHandler
from typing import Any, Dict, List, Optional, Tuple

import requests

from ..codex_instructions import get_codex_instructions
from ..codex_oauth import (
    CodexOAuthTokens,
    extract_chatgpt_account_id,
    load_tokens,
    refresh_access_token,
    save_tokens,
)
from ..config import Settings
from ..context_limits import enforce_context_limits
from ..logging_utils import log_payload
from ..model_registry import default_extra_body_for
from ..streaming import stream_responses_response

logger = logging.getLogger(__name__)

OPENAI_BETA_HEADER = "OpenAI-Beta"
OPENAI_BETA_VALUE = "responses=experimental"
CHATGPT_ACCOUNT_HEADER = "chatgpt-account-id"
ORIGINATOR_HEADER = "originator"
ORIGINATOR_VALUE = "codex_cli_rs"

_CODEX_GLOBAL_DEFAULTS: Dict[str, Any] = {
    "reasoning": {"effort": "medium", "summary": "auto"},
    "text": {"verbosity": "medium"},
    "include": ["reasoning.encrypted_content"],
    "store": False,
}


def _codex_model_key(settings: Settings, target_model: str) -> str:
    selected = (getattr(settings, "model", "") or "").strip()
    if selected.lower().startswith("codex:"):
        _, name = selected.split(":", 1)
        if name.strip():
            return f"codex:{name.strip()}"
    return f"codex:{target_model}"


def _tokens_from_settings(settings: Settings) -> Optional[CodexOAuthTokens]:
    access = getattr(settings, "codex_access_token", "") or ""
    refresh = getattr(settings, "codex_refresh_token", "") or ""
    expires = int(getattr(settings, "codex_expires_at_ms", 0) or 0)
    if access and refresh and expires > 0:
        return CodexOAuthTokens(access=access, refresh=refresh, expires_at_ms=expires)
    return None


def _normalized_auth_mode(settings: Settings) -> str:
    mode = str(getattr(settings, "codex_auth", "") or "").strip().lower()
    if mode in {"oauth", "login", "file", "stored"}:
        return "oauth"
    if mode in {"env", "token", "tokens"}:
        return "env"
    return "auto"


def _apply_tokens_to_settings(settings: Settings, tokens: CodexOAuthTokens) -> None:
    try:
        settings.codex_access_token = tokens.access
        settings.codex_refresh_token = tokens.refresh
        settings.codex_expires_at_ms = tokens.expires_at_ms
    except Exception:
        return


def _resolve_codex_auth(settings: Settings) -> Tuple[CodexOAuthTokens, str]:
    auth_mode = _normalized_auth_mode(settings)
    tokens: Optional[CodexOAuthTokens] = None
    from_file = False
    if auth_mode in {"auto", "env"}:
        tokens = _tokens_from_settings(settings)
    if not tokens and auth_mode in {"auto", "oauth"}:
        tokens = load_tokens()
        from_file = True
    if not tokens:
        if auth_mode == "env":
            raise RuntimeError(
                "OpenAI Codex OAuth env tokens not configured (set OPENAI_CODEX_ACCESS_TOKEN/OPENAI_CODEX_REFRESH_TOKEN/OPENAI_CODEX_EXPIRES_AT_MS)"
            )
        raise RuntimeError("OpenAI Codex OAuth not configured (run cc-adapter-codex-login)")

    if tokens.expired():
        tokens = refresh_access_token(
            tokens.refresh,
            proxies=settings.resolved_proxies(),
            timeout=float(settings.lmstudio_timeout),
        )
        _apply_tokens_to_settings(settings, tokens)
        if from_file:
            save_tokens(tokens)
    else:
        _apply_tokens_to_settings(settings, tokens)

    account_id = extract_chatgpt_account_id(tokens.access)
    if not account_id:
        raise RuntimeError("Failed to extract chatgpt_account_id from OAuth token")
    return tokens, account_id


def _headers(account_id: str, access_token: str) -> Dict[str, str]:
    return {
        "Authorization": f"Bearer {access_token}",
        CHATGPT_ACCOUNT_HEADER: account_id,
        OPENAI_BETA_HEADER: OPENAI_BETA_VALUE,
        ORIGINATOR_HEADER: ORIGINATOR_VALUE,
        "accept": "text/event-stream",
    }


def _responses_tools(tools: Any) -> Optional[List[Dict[str, Any]]]:
    if not isinstance(tools, list):
        return None
    out: List[Dict[str, Any]] = []
    for tool in tools:
        if not isinstance(tool, dict):
            continue
        if tool.get("type") != "function":
            continue
        func = tool.get("function") or {}
        if not isinstance(func, dict):
            continue
        name = func.get("name")
        if not name:
            continue
        entry: Dict[str, Any] = {"type": "function", "name": str(name)}
        desc = func.get("description")
        if desc:
            entry["description"] = str(desc)
        params = func.get("parameters")
        if params is not None:
            entry["parameters"] = params
        out.append(entry)
    return out or None


def _responses_tool_choice(tool_choice: Any) -> Any:
    if isinstance(tool_choice, str):
        return tool_choice
    if isinstance(tool_choice, dict):
        if tool_choice.get("type") == "function":
            fn = tool_choice.get("function") or {}
            if isinstance(fn, dict) and fn.get("name"):
                return {"type": "function", "name": fn.get("name")}
            if tool_choice.get("name"):
                return {"type": "function", "name": tool_choice.get("name")}
    return tool_choice


def _responses_content_parts(
    content: Any,
    *,
    text_part_type: str = "input_text",
    include_images: bool = True,
) -> List[Dict[str, Any]]:
    parts: List[Dict[str, Any]] = []
    if content is None:
        return parts
    if isinstance(content, str):
        if content:
            parts.append({"type": text_part_type, "text": content})
        return parts
    if isinstance(content, list):
        for item in content:
            if isinstance(item, str):
                if item:
                    parts.append({"type": text_part_type, "text": item})
                continue
            if not isinstance(item, dict):
                continue
            item_type = item.get("type")
            if item_type in ("text", "input_text", "output_text") and item.get("text") is not None:
                text = str(item.get("text") or "")
                if text:
                    parts.append({"type": text_part_type, "text": text})
                continue
            if item_type == "image_url":
                if not include_images:
                    continue
                image_url = item.get("image_url") or {}
                if isinstance(image_url, dict):
                    url = image_url.get("url")
                else:
                    url = image_url
                if url:
                    parts.append({"type": "input_image", "image_url": str(url)})
                continue
            if "text" in item and item.get("text") is not None:
                text = str(item.get("text") or "")
                if text:
                    parts.append({"type": text_part_type, "text": text})
    if isinstance(content, dict) and "text" in content:
        text = str(content.get("text") or "")
        if text:
            parts.append({"type": text_part_type, "text": text})
    return parts


def _flatten_text(content: Any) -> str:
    if content is None:
        return ""
    if isinstance(content, str):
        return content
    if isinstance(content, list):
        bits: List[str] = []
        for item in content:
            if isinstance(item, str):
                bits.append(item)
            elif isinstance(item, dict):
                if "text" in item and item.get("text") is not None:
                    bits.append(str(item.get("text") or ""))
        return "\n".join([b for b in bits if b])
    if isinstance(content, dict) and "text" in content:
        return str(content.get("text") or "")
    return str(content)


def _messages_to_responses_input(messages: List[Dict[str, Any]]) -> Tuple[str, List[Dict[str, Any]]]:
    developer_parts: List[str] = []
    items: List[Dict[str, Any]] = []

    for msg in messages:
        if not isinstance(msg, dict):
            continue
        role = (msg.get("role") or "").strip()
        if role == "system":
            text = _flatten_text(msg.get("content"))
            if text:
                developer_parts.append(text)
            continue

        if role == "tool":
            call_id = msg.get("tool_call_id") or msg.get("id")
            output = _flatten_text(msg.get("content"))
            if call_id:
                items.append(
                    {
                        "type": "function_call_output",
                        "call_id": str(call_id),
                        "output": output,
                    }
                )
            else:
                # Best-effort fallback: surface tool output as text context.
                if output:
                    items.append(
                        {
                            "type": "message",
                            "role": "user",
                            "content": [{"type": "input_text", "text": output}],
                        }
                    )
            continue

        content_parts = _responses_content_parts(
            msg.get("content"),
            text_part_type="output_text" if role == "assistant" else "input_text",
            include_images=(role != "assistant"),
        )
        if content_parts:
            items.append({"type": "message", "role": role or "user", "content": content_parts})

        if role == "assistant":
            for call in msg.get("tool_calls") or []:
                if not isinstance(call, dict):
                    continue
                func = call.get("function") or {}
                if not isinstance(func, dict):
                    continue
                call_id = call.get("id") or func.get("name") or "tool_call"
                name = func.get("name") or "tool"
                arguments = func.get("arguments") or ""
                items.append(
                    {
                        "type": "function_call",
                        "call_id": str(call_id),
                        "name": str(name),
                        "arguments": str(arguments),
                    }
                )

    developer_prompt = "\n\n".join([p for p in developer_parts if p]).strip()
    return developer_prompt, items


def _parse_final_response(resp: requests.Response) -> Dict[str, Any]:
    for line in resp.iter_lines(decode_unicode=False):
        if not line:
            continue
        if not line.startswith(b"data:"):
            continue
        raw = line[len(b"data:") :].strip()
        if raw == b"[DONE]":
            break
        try:
            event_obj = json.loads(raw.decode("utf-8"))
        except Exception:
            continue
        etype = str(event_obj.get("type") or "")
        if etype in ("response.completed", "response.done") and isinstance(event_obj.get("response"), dict):
            return event_obj["response"]
        if etype == "error":
            raise RuntimeError(str(event_obj.get("message") or event_obj.get("error") or "Unknown error"))
    raise RuntimeError("Codex backend stream ended without a final response event")


def _responses_to_chat_completions(response_obj: Dict[str, Any]) -> Dict[str, Any]:
    output_items = response_obj.get("output") or []
    tool_calls: List[Dict[str, Any]] = []
    text_bits: List[str] = []

    if isinstance(response_obj.get("output_text"), str) and response_obj.get("output_text"):
        text_bits.append(str(response_obj.get("output_text")))

    if isinstance(output_items, list):
        for item in output_items:
            if not isinstance(item, dict):
                continue
            itype = item.get("type")
            if itype == "message":
                content = item.get("content")
                if isinstance(content, str):
                    if content:
                        text_bits.append(content)
                elif isinstance(content, list):
                    for part in content:
                        if isinstance(part, dict) and part.get("text") is not None:
                            maybe = str(part.get("text") or "")
                            if maybe:
                                text_bits.append(maybe)
            elif itype == "function_call":
                call_id = item.get("call_id") or item.get("id") or "tool_call"
                name = item.get("name") or "tool"
                args = item.get("arguments") or ""
                tool_calls.append(
                    {
                        "id": str(call_id),
                        "type": "function",
                        "function": {"name": str(name), "arguments": str(args)},
                    }
                )

    text = "\n".join([t for t in text_bits if t]).strip()
    message: Dict[str, Any] = {"role": "assistant", "content": [{"type": "text", "text": text}] if text else ""}
    if tool_calls:
        message["tool_calls"] = tool_calls

    finish_reason = "tool_calls" if tool_calls else "stop"
    chat: Dict[str, Any] = {
        "id": response_obj.get("id") or "resp",
        "choices": [{"index": 0, "message": message, "finish_reason": finish_reason}],
    }
    usage = response_obj.get("usage") or {}
    if isinstance(usage, dict):
        mapped: Dict[str, Any] = {}
        if "input_tokens" in usage:
            mapped["prompt_tokens"] = usage.get("input_tokens")
        if "output_tokens" in usage:
            mapped["completion_tokens"] = usage.get("output_tokens")
        if mapped:
            chat["usage"] = mapped
    return chat


def _request_body(
    payload: Dict[str, Any],
    settings: Settings,
    *,
    model_key: Optional[str] = None,
    force_refresh_instructions: bool = False,
) -> Dict[str, Any]:
    developer_prompt, input_items = _messages_to_responses_input(payload.get("messages") or [])
    model_defaults = default_extra_body_for(model_key) if model_key else {}

    timeout = min(30.0, float(settings.lmstudio_timeout))
    instructions = get_codex_instructions(
        payload.get("model") or "",
        proxies=settings.resolved_proxies(),
        timeout=timeout,
        force_refresh=force_refresh_instructions,
    )

    if developer_prompt:
        input_items = [
            {
                "type": "message",
                "role": "developer",
                "content": [{"type": "input_text", "text": developer_prompt}],
            },
            *input_items,
        ]

    body: Dict[str, Any] = {
        "model": payload.get("model") or "",
        "store": False,
        "stream": True,
        "instructions": instructions,
        "input": input_items,
        "include": ["reasoning.encrypted_content"],
    }

    if isinstance(model_defaults.get("include"), list) and model_defaults.get("include"):
        body["include"] = model_defaults.get("include")
    if isinstance(payload.get("include"), list) and payload.get("include"):
        body["include"] = payload.get("include")

    # ChatGPT Codex backend rejects sampling parameters such as temperature/top_p.

    reasoning = copy.deepcopy(_CODEX_GLOBAL_DEFAULTS.get("reasoning") or {})
    if isinstance(payload.get("reasoning"), dict) and payload.get("reasoning"):
        reasoning.update(payload.get("reasoning"))
    if isinstance(model_defaults.get("reasoning"), dict) and model_defaults.get("reasoning"):
        # Preset defaults override per-request hints.
        reasoning.update(model_defaults.get("reasoning"))
    if reasoning:
        body["reasoning"] = reasoning

    text = copy.deepcopy(_CODEX_GLOBAL_DEFAULTS.get("text") or {})
    if isinstance(payload.get("text"), dict) and payload.get("text"):
        text.update(payload.get("text"))
    if isinstance(model_defaults.get("text"), dict) and model_defaults.get("text"):
        # Preset defaults override per-request hints.
        text.update(model_defaults.get("text"))
    if text:
        body["text"] = text

    tools = _responses_tools(payload.get("tools"))
    if tools:
        body["tools"] = tools

    tool_choice = payload.get("tool_choice")
    if tool_choice is not None:
        body["tool_choice"] = _responses_tool_choice(tool_choice)

    return body


def send(payload: Dict[str, Any], settings: Settings, target_model: str) -> Dict[str, Any]:
    tokens, account_id = _resolve_codex_auth(settings)
    model_key = _codex_model_key(settings, target_model)
    trimmed_payload, trim_meta = enforce_context_limits(payload, settings, model_key)
    if trim_meta.get("dropped"):
        logger.warning(
            "Trimmed %s message(s) for Codex context (est %s -> %s tokens, budget=%s)",
            trim_meta["dropped"],
            trim_meta.get("before", 0),
            trim_meta.get("after", 0),
            trim_meta.get("budget", 0),
        )

    req_payload = dict(trimmed_payload)
    req_payload["model"] = target_model
    body = _request_body(req_payload, settings, model_key=model_key)
    log_payload(logger, f"Codex request -> {target_model}", body)

    resp = requests.post(
        settings.codex_base_url,
        json=body,
        headers=_headers(account_id, tokens.access),
        timeout=float(settings.lmstudio_timeout),
        proxies=settings.resolved_proxies(),
        stream=True,
    )
    try:
        resp.raise_for_status()
    except requests.HTTPError as exc:
        body_text = resp.text or ""
        if resp.status_code == 400 and "Instructions are not valid" in body_text:
            try:
                resp.close()
            except Exception:
                pass
            logger.warning("Codex backend rejected instructions; refreshing and retrying once.")
            body = _request_body(req_payload, settings, model_key=model_key, force_refresh_instructions=True)
            resp = requests.post(
                settings.codex_base_url,
                json=body,
                headers=_headers(account_id, tokens.access),
                timeout=float(settings.lmstudio_timeout),
                proxies=settings.resolved_proxies(),
                stream=True,
            )
            try:
                resp.raise_for_status()
            except requests.HTTPError as exc2:
                raise requests.HTTPError(f"{exc2} | body={resp.text}") from exc2
        else:
            raise requests.HTTPError(f"{exc} | body={body_text}") from exc

    try:
        final = _parse_final_response(resp)
        chat = _responses_to_chat_completions(final)
        log_payload(logger, "Codex final response (parsed)", final)
        return chat
    finally:
        try:
            resp.close()
        except Exception:
            pass


def stream(
    payload: Dict[str, Any],
    settings: Settings,
    requested_model: str,
    incoming: Optional[Dict[str, Any]],
    handler: BaseHTTPRequestHandler,
    logger,
):
    tokens, account_id = _resolve_codex_auth(settings)
    model_key = _codex_model_key(settings, requested_model)
    trimmed_payload, trim_meta = enforce_context_limits(payload, settings, model_key)
    if trim_meta.get("dropped"):
        logger.warning(
            "Trimmed %s message(s) for Codex context (est %s -> %s tokens, budget=%s)",
            trim_meta["dropped"],
            trim_meta.get("before", 0),
            trim_meta.get("after", 0),
            trim_meta.get("budget", 0),
        )

    req_payload = dict(trimmed_payload)
    req_payload["model"] = requested_model
    body = _request_body(req_payload, settings, model_key=model_key)
    log_payload(logger, f"Codex stream request -> {requested_model}", body)

    resp = requests.post(
        settings.codex_base_url,
        json=body,
        headers=_headers(account_id, tokens.access),
        timeout=float(settings.lmstudio_timeout),
        proxies=settings.resolved_proxies(),
        stream=True,
    )
    try:
        resp.raise_for_status()
    except requests.HTTPError as exc:
        body_text = resp.text or ""
        if resp.status_code == 400 and "Instructions are not valid" in body_text:
            try:
                resp.close()
            except Exception:
                pass
            logger.warning("Codex backend rejected instructions; refreshing and retrying once.")
            body = _request_body(req_payload, settings, model_key=model_key, force_refresh_instructions=True)
            resp = requests.post(
                settings.codex_base_url,
                json=body,
                headers=_headers(account_id, tokens.access),
                timeout=float(settings.lmstudio_timeout),
                proxies=settings.resolved_proxies(),
                stream=True,
            )
            try:
                resp.raise_for_status()
            except requests.HTTPError as exc2:
                raise requests.HTTPError(f"{exc2} | body={resp.text}") from exc2
        else:
            raise requests.HTTPError(f"{exc} | body={body_text}") from exc

    handler.send_response(200)
    handler.send_header("Content-Type", "text/event-stream")
    handler.send_header("Cache-Control", "no-cache")
    handler.send_header("Connection", "keep-alive")
    handler.end_headers()

    stream_responses_response(resp, requested_model, incoming, handler, logger)



import unittest

from cc_adapter.providers import poe


class PoePayloadMergeTestCase(unittest.TestCase):
    def test_merge_adds_thinking_budget_without_overwriting_extra_body(self):
        payload = {"messages": [], "extra_body": {"foo": "bar"}}
        incoming = {"thinking": {"budget_tokens": 256}}

        merged = poe._merge_extra_body(payload, incoming, defaults={"web_search": True})

        self.assertIn("extra_body", merged)
        self.assertEqual(merged["extra_body"]["foo"], "bar")
        self.assertEqual(merged["extra_body"].get("thinking_budget"), 256)
        self.assertTrue(merged["extra_body"].get("web_search"))

    def test_merge_sets_web_search_only_when_requested(self):
        merged = poe._merge_extra_body({}, {}, defaults={})
        self.assertNotIn("extra_body", merged)

        incoming = {
            "tools": [
                {"function": {"name": "web_search"}, "description": "", "input_schema": {"type": "object"}}
            ]
        }
        merged = poe._merge_extra_body({}, incoming, defaults={})
        self.assertTrue(merged["extra_body"].get("web_search"))

    def test_merge_preserves_explicit_web_search(self):
        payload = {"extra_body": {"web_search": False}}
        incoming = {
            "tools": [
                {"function": {"name": "web_search"}, "description": "", "input_schema": {"type": "object"}}
            ]
        }

        merged = poe._merge_extra_body(payload, incoming, defaults={"web_search": True})

        self.assertEqual(merged["extra_body"].get("web_search"), False)

    def test_default_web_search_applied_only_to_claude(self):
        claude_defaults = {"web_search": True}
        merged_claude = poe._merge_extra_body({}, {}, defaults=claude_defaults)
        self.assertTrue(merged_claude["extra_body"].get("web_search"))

        deepseek_defaults = {}
        merged_deepseek = poe._merge_extra_body({}, {}, defaults=deepseek_defaults)
        self.assertNotIn("extra_body", merged_deepseek)


if __name__ == "__main__":
    unittest.main()



import json
import os
import tempfile
import unittest
from pathlib import Path
from unittest import mock

from cc_adapter.config import Settings
from cc_adapter.codex_oauth import DEFAULT_TOKEN_FILENAME
from cc_adapter.models import available_models


class CodexModelListingTestCase(unittest.TestCase):
    def test_available_models_includes_codex_when_oauth_file_present(self):
        with tempfile.TemporaryDirectory() as cfgdir:
            with mock.patch.dict(os.environ, {"CC_ADAPTER_CONFIG_DIR": cfgdir}, clear=False):
                Path(cfgdir, DEFAULT_TOKEN_FILENAME).write_text(
                    json.dumps(
                        {
                            "type": "oauth",
                            "access": "a",
                            "refresh": "r",
                            "expires_at_ms": 9_999_999_999_999,
                        }
                    ),
                    encoding="utf-8",
                )
                settings = Settings(model="lmstudio:local", lmstudio_model="local", codex_auth="oauth")
                models = available_models(settings)
                self.assertIn("codex:gpt-5.1-codex", models)
                self.assertIn("codex:gpt-5.2-medium", models)
                self.assertIn("codex:gpt-5.1-codex-max-xhigh", models)

    def test_available_models_hides_codex_when_env_mode_and_no_env_tokens(self):
        with tempfile.TemporaryDirectory() as cfgdir:
            with mock.patch.dict(os.environ, {"CC_ADAPTER_CONFIG_DIR": cfgdir}, clear=False):
                Path(cfgdir, DEFAULT_TOKEN_FILENAME).write_text(
                    json.dumps(
                        {
                            "type": "oauth",
                            "access": "a",
                            "refresh": "r",
                            "expires_at_ms": 9_999_999_999_999,
                        }
                    ),
                    encoding="utf-8",
                )
                settings = Settings(model="lmstudio:local", lmstudio_model="local", codex_auth="env")
                models = available_models(settings)
                self.assertNotIn("codex:gpt-5.1-codex", models)


if __name__ == "__main__":
    unittest.main()



import base64
import json
import unittest

from cc_adapter import codex_oauth


def _jwt(payload: dict) -> str:
    header = {"alg": "none", "typ": "JWT"}

    def b64url(obj: dict) -> str:
        return base64.urlsafe_b64encode(json.dumps(obj).encode("utf-8")).decode("utf-8").rstrip("=")

    return f"{b64url(header)}.{b64url(payload)}.sig"


class CodexOAuthTestCase(unittest.TestCase):
    def test_parse_authorization_input_url(self):
        code, state = codex_oauth.parse_authorization_input(
            "http://localhost:1455/auth/callback?code=abc123&state=xyz789"
        )
        self.assertEqual(code, "abc123")
        self.assertEqual(state, "xyz789")

    def test_parse_authorization_input_code_hash_state(self):
        code, state = codex_oauth.parse_authorization_input("abc123#xyz789")
        self.assertEqual(code, "abc123")
        self.assertEqual(state, "xyz789")

    def test_parse_authorization_input_query_string(self):
        code, state = codex_oauth.parse_authorization_input("code=abc123&state=xyz789")
        self.assertEqual(code, "abc123")
        self.assertEqual(state, "xyz789")

    def test_parse_authorization_input_code_only(self):
        code, state = codex_oauth.parse_authorization_input("abc123")
        self.assertEqual(code, "abc123")
        self.assertIsNone(state)

    def test_parse_authorization_input_empty(self):
        code, state = codex_oauth.parse_authorization_input("")
        self.assertIsNone(code)
        self.assertIsNone(state)

    def test_extract_chatgpt_account_id(self):
        token = _jwt(
            {
                codex_oauth.JWT_CLAIM_PATH: {
                    codex_oauth.JWT_ACCOUNT_ID_FIELD: "account-123",
                }
            }
        )
        self.assertEqual(codex_oauth.extract_chatgpt_account_id(token), "account-123")


if __name__ == "__main__":
    unittest.main()




import copy
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional, Tuple


@dataclass(frozen=True)
class ModelInfo:
    """Metadata for a provider model."""

    provider: str
    slug: str
    upstream: Optional[str] = None
    aliases: Tuple[str, ...] = ()
    context_window: int = 0
    priority: int = 100
    expose: bool = True
    extra_body: Dict[str, Any] = field(default_factory=dict)

    @property
    def target(self) -> str:
        """Return the provider-facing model identifier."""
        return self.upstream or self.slug


# Ordering favors higher-cost/capability models first so dropdowns and /v1/models
# responses present sensible defaults.
MODEL_ENTRIES: Tuple[ModelInfo, ...] = (
    # LM Studio
    ModelInfo(
        provider="lmstudio",
        slug="gpt-oss-120b",
        context_window=131_072,
        priority=60,
    ),
    # OpenAI Codex (ChatGPT OAuth)
    ModelInfo(
        provider="codex",
        slug="gpt-5.2",
        context_window=400_000,
        priority=5,
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-max",
        aliases=("gpt-5-codex-max",),
        context_window=400_000,
        priority=10,
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex",
        aliases=("gpt-5-codex",),
        context_window=400_000,
        priority=20,
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-mini",
        aliases=("gpt-5-codex-mini", "codex-mini-latest"),
        context_window=400_000,
        priority=30,
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1",
        aliases=("gpt-5",),
        context_window=400_000,
        priority=40,
    ),
    # Codex presets (mirrors ref/opencode-openai-codex-auth/config/full-opencode.json)
    ModelInfo(
        provider="codex",
        slug="gpt-5.2-low",
        upstream="gpt-5.2",
        context_window=400_000,
        priority=6,
        extra_body={
            "reasoning": {"effort": "low", "summary": "auto"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.2-medium",
        upstream="gpt-5.2",
        context_window=400_000,
        priority=7,
        extra_body={
            "reasoning": {"effort": "medium", "summary": "auto"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.2-high",
        upstream="gpt-5.2",
        context_window=400_000,
        priority=8,
        extra_body={
            "reasoning": {"effort": "high", "summary": "detailed"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.2-xhigh",
        upstream="gpt-5.2",
        context_window=400_000,
        priority=9,
        extra_body={
            "reasoning": {"effort": "xhigh", "summary": "detailed"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-max-low",
        upstream="gpt-5.1-codex-max",
        context_window=400_000,
        priority=11,
        extra_body={
            "reasoning": {"effort": "low", "summary": "detailed"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-max-medium",
        upstream="gpt-5.1-codex-max",
        context_window=400_000,
        priority=12,
        extra_body={
            "reasoning": {"effort": "medium", "summary": "detailed"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-max-high",
        upstream="gpt-5.1-codex-max",
        context_window=400_000,
        priority=13,
        extra_body={
            "reasoning": {"effort": "high", "summary": "detailed"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-max-xhigh",
        upstream="gpt-5.1-codex-max",
        context_window=400_000,
        priority=14,
        extra_body={
            "reasoning": {"effort": "xhigh", "summary": "detailed"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-low",
        upstream="gpt-5.1-codex",
        context_window=400_000,
        priority=21,
        extra_body={
            "reasoning": {"effort": "low", "summary": "auto"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-medium",
        upstream="gpt-5.1-codex",
        context_window=400_000,
        priority=22,
        extra_body={
            "reasoning": {"effort": "medium", "summary": "auto"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-high",
        upstream="gpt-5.1-codex",
        context_window=400_000,
        priority=23,
        extra_body={
            "reasoning": {"effort": "high", "summary": "detailed"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-mini-medium",
        upstream="gpt-5.1-codex-mini",
        context_window=400_000,
        priority=31,
        extra_body={
            "reasoning": {"effort": "medium", "summary": "auto"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-codex-mini-high",
        upstream="gpt-5.1-codex-mini",
        context_window=400_000,
        priority=32,
        extra_body={
            "reasoning": {"effort": "high", "summary": "detailed"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-low",
        upstream="gpt-5.1",
        context_window=400_000,
        priority=41,
        extra_body={
            "reasoning": {"effort": "low", "summary": "auto"},
            "text": {"verbosity": "low"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-medium",
        upstream="gpt-5.1",
        context_window=400_000,
        priority=42,
        extra_body={
            "reasoning": {"effort": "medium", "summary": "auto"},
            "text": {"verbosity": "medium"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    ModelInfo(
        provider="codex",
        slug="gpt-5.1-high",
        upstream="gpt-5.1",
        context_window=400_000,
        priority=43,
        extra_body={
            "reasoning": {"effort": "high", "summary": "detailed"},
            "text": {"verbosity": "high"},
            "include": ["reasoning.encrypted_content"],
            "store": False,
        },
    ),
    # Poe
    ModelInfo(
        provider="poe",
        slug="claude-opus-4.5",
        aliases=("claude-opus-4-5",),
        context_window=200_000,
        priority=10,
        extra_body={"web_search": True},
    ),
    ModelInfo(
        provider="poe",
        slug="claude-sonnet-4.5",
        aliases=("claude-sonnet-4-5",),
        context_window=1_000_000,
        priority=20,
        extra_body={"web_search": True},
    ),
    ModelInfo(
        provider="poe",
        slug="claude-haiku-4.5",
        aliases=("claude-haiku-4-5", "claude-haiku-4-5-20251001"),
        context_window=200_000,
        priority=30,
        extra_body={"web_search": True},
    ),
    ModelInfo(
        provider="poe",
        slug="deepseek-v3.2",
        context_window=163_840,
        priority=40,
    ),
    ModelInfo(
        provider="poe",
        slug="glm-4.6",
        context_window=202_752,
        priority=50,
    ),
    ModelInfo(
        provider="poe",
        slug="gpt-5.2-pro",
        context_window=400_000,
        priority=99,
        extra_body={"web_search": True},
    ),
    # OpenRouter
    ModelInfo(
        provider="openrouter",
        slug="claude-opus-4.5",
        upstream="anthropic/claude-opus-4.5",
        context_window=200_000,
        priority=10,
    ),
    ModelInfo(
        provider="openrouter",
        slug="claude-sonnet-4.5",
        upstream="anthropic/claude-sonnet-4.5",
        context_window=1_000_000,
        priority=20,
    ),
    ModelInfo(
        provider="openrouter",
        slug="claude-haiku-4.5",
        upstream="anthropic/claude-haiku-4.5",
        aliases=("claude-haiku-4-5", "claude-haiku-4-5-20251001"),
        context_window=200_000,
        priority=30,
    ),
    ModelInfo(
        provider="openrouter",
        slug="gpt-5.2",
        upstream="openai/gpt-5.2",
        context_window=400_000,
        priority=40,
    ),
    ModelInfo(
        provider="openrouter",
        slug="glm-4.6",
        upstream="z-ai/glm-4.6",
        context_window=202_752,
        priority=50,
    ),
    ModelInfo(
        provider="openrouter",
        slug="gpt-5.2-pro",
        upstream="openai/gpt-5.2-pro",
        context_window=400_000,
        priority=99,
    ),
)

DEFAULT_PROVIDER_MODELS: Dict[str, str] = {
    "lmstudio": "gpt-oss-120b",
    "codex": "gpt-5.1-codex",
    "poe": "claude-opus-4.5",
    "openrouter": "claude-opus-4.5",
}


def _normalize(name: str) -> str:
    return name.strip().lower()


def _names_for(info: ModelInfo) -> List[str]:
    names: List[str] = [info.slug]
    if info.upstream:
        names.append(info.upstream)
    names.extend(info.aliases)
    return [n for n in names if n]


def _build_provider_lookup(entries: Iterable[ModelInfo]) -> Dict[str, Dict[str, ModelInfo]]:
    lookup: Dict[str, Dict[str, ModelInfo]] = {}
    for info in entries:
        provider_map = lookup.setdefault(info.provider, {})
        for name in _names_for(info):
            provider_map.setdefault(_normalize(name), info)
    return lookup


def _build_global_lookup(entries: Iterable[ModelInfo]) -> Dict[str, ModelInfo]:
    # Earlier priorities win ties when aliases collide across providers.
    ordered = sorted(entries, key=lambda i: (i.priority, i.provider, i.slug))
    lookup: Dict[str, ModelInfo] = {}
    for info in ordered:
        for name in _names_for(info):
            key = _normalize(name)
            lookup.setdefault(key, info)
    return lookup


_PROVIDER_MODELS: Dict[str, List[ModelInfo]] = {}
for info in MODEL_ENTRIES:
    _PROVIDER_MODELS.setdefault(info.provider, []).append(info)
for provider in _PROVIDER_MODELS:
    _PROVIDER_MODELS[provider].sort(key=lambda i: (i.priority, i.slug))

_PROVIDER_LOOKUP: Dict[str, Dict[str, ModelInfo]] = _build_provider_lookup(MODEL_ENTRIES)
_GLOBAL_LOOKUP: Dict[str, ModelInfo] = _build_global_lookup(MODEL_ENTRIES)


def provider_models(provider: str, expose_only: bool = True) -> List[ModelInfo]:
    """Return known models for a provider, optionally hiding non-exposed entries."""
    models = _PROVIDER_MODELS.get(provider.lower(), [])
    if expose_only:
        return [m for m in models if m.expose]
    return list(models)


def provider_model_slugs(provider: str, expose_only: bool = True) -> List[str]:
    """List provider models as user-facing slugs (no provider prefix)."""
    return [m.slug for m in provider_models(provider, expose_only=expose_only)]


def find_model(provider: Optional[str], name: str) -> Optional[ModelInfo]:
    """Find a model definition by provider/alias/slug."""
    normalized = _normalize(name)
    if provider:
        return _PROVIDER_LOOKUP.get(provider.lower(), {}).get(normalized)
    return _GLOBAL_LOOKUP.get(normalized)


def canonicalize_model(provider: str, name: str) -> str:
    """Normalize a provider/model string to the upstream identifier."""
    info = find_model(provider, name)
    if info:
        return info.target
    fallback = find_model(None, name)
    return fallback.target if fallback else name


def default_context_window_for(model: Optional[str]) -> int:
    """Return a sensible default context window for known models."""
    if not model:
        return 0
    provider = ""
    name = model
    if ":" in model:
        provider, name = model.split(":", 1)
        provider = provider.lower()

    info = find_model(provider or None, name)
    if info:
        return info.context_window
    return 0


def default_extra_body_for(model: Optional[str]) -> Dict[str, Any]:
    """Return provider-specific default extra_body payload settings, if any."""
    if not model:
        return {}
    provider = ""
    name = model
    if ":" in model:
        provider, name = model.split(":", 1)
        provider = provider.lower()

    info = find_model(provider or None, name)
    if info and info.extra_body:
        return copy.deepcopy(info.extra_body)
    return {}



import unittest

from cc_adapter.models import normalize_model_spec


class ModelSpecNormalizationTestCase(unittest.TestCase):
    def test_normalize_model_spec_expands_provider_only(self):
        self.assertEqual(normalize_model_spec("codex:"), "codex:gpt-5.1-codex")
        self.assertEqual(normalize_model_spec("codex"), "codex:gpt-5.1-codex")
        self.assertEqual(normalize_model_spec("poe:"), "poe:claude-opus-4.5")
        self.assertEqual(normalize_model_spec("openrouter:"), "openrouter:claude-opus-4.5")
        self.assertEqual(normalize_model_spec("lmstudio:"), "lmstudio:gpt-oss-120b")

    def test_normalize_model_spec_keeps_fully_qualified(self):
        self.assertEqual(normalize_model_spec("codex:gpt-5.2-high"), "codex:gpt-5.2-high")
        self.assertEqual(normalize_model_spec("poe:deepseek-v3.2"), "poe:deepseek-v3.2")


if __name__ == "__main__":
    unittest.main()




#!/usr/bin/env python3
"""
Main adapter server entrypoint.
"""

import argparse
import json
import logging
import os
import sys
import threading
from dataclasses import replace
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
import socket
from typing import Any, Dict, Optional
from urllib.parse import urlparse
from subprocess import Popen, DEVNULL

from .config import Settings, load_settings, apply_overrides
from .models import available_models, normalize_model_spec, resolve_provider_model
from .converters import anthropic_to_openai, openai_to_anthropic
from .providers import lmstudio, poe, openrouter, codex
from . import streaming
from .logging_utils import log_payload, resolve_log_level

logging.basicConfig(
    level=resolve_log_level(os.getenv("LOG_LEVEL", "INFO")),
    format="%(asctime)s %(levelname)s %(message)s",
)
logger = logging.getLogger("cc-adapter")

CODEX_HAIKU_FALLBACK_MODEL = "gpt-5.1-codex-mini"

def _effective_codex_settings(
    settings: Settings,
    requested_model: Any,
    upstream_model: str,
) -> tuple[Settings, str]:
    requested_text = str(requested_model or "")
    if not requested_text or "claude-haiku" not in requested_text.lower():
        return settings, upstream_model
    if not str(getattr(settings, "model", "") or "").lower().startswith("codex:"):
        return settings, upstream_model
    return replace(settings, model=f"codex:{CODEX_HAIKU_FALLBACK_MODEL}"), CODEX_HAIKU_FALLBACK_MODEL


def port_available(host: str, port: int) -> bool:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(1.0)
    try:
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        except Exception:
            pass
        sock.bind((host, port))
        return True
    except OSError:
        return False
    finally:
        sock.close()


def _json_response(handler: BaseHTTPRequestHandler, status: int, payload: Dict[str, Any]):
    body = json.dumps(payload).encode("utf-8")
    try:
        handler.send_response(status)
        handler.send_header("Content-Type", "application/json")
        handler.send_header("Content-Length", str(len(body)))
        handler.end_headers()
        handler.wfile.write(body)
    except (BrokenPipeError, ConnectionResetError):
        logger.info("Client disconnected before response could be sent (status=%s)", status)
        handler.close_connection = True
    except Exception:
        handler.close_connection = True
        logger.exception("Failed to send JSON response")


class AdapterHTTPServer(ThreadingHTTPServer):
    """HTTP server that suppresses noisy client disconnect tracebacks."""

    def handle_error(self, request, client_address):
        exc_type, exc, _ = sys.exc_info()
        if exc_type in (ConnectionResetError, BrokenPipeError):
            logger.debug("Client disconnected during request from %s", client_address)
            return
        return super().handle_error(request, client_address)


class AdapterHandler(BaseHTTPRequestHandler):
    settings: Settings = load_settings()

    def log_message(self, format: str, *args: Any) -> None:
        logger.info("%s - %s", self.client_address[0], format % args)

    def _client_is_loopback(self) -> bool:
        host = str((self.client_address[0] or "")).strip()
        return host == "127.0.0.1" or host == "::1" or host.startswith("127.")

    def do_GET(self):
        parsed = urlparse(self.path)
        if parsed.path == "/health":
            return _json_response(self, 200, {"status": "ok"})
        if parsed.path == "/v1/models":
            return _json_response(
                self,
                200,
                {
                    "data": [{"id": m, "object": "model"} for m in available_models(self.settings)],
                },
            )
        if parsed.path == "/v1/messages/count_tokens":
            # Lightweight token estimate based on character counts.
            length = int(self.headers.get("Content-Length", "0"))
            raw_body = self.rfile.read(length) if length else b"{}"
            try:
                incoming = json.loads(raw_body.decode("utf-8") or "{}")
            except Exception as exc:
                logger.exception("Failed to parse count_tokens request")
                return _json_response(self, 400, {"error": f"Invalid JSON: {exc}"})
            prompt_tokens = streaming.estimate_prompt_tokens(incoming)
            return _json_response(self, 200, {"input_tokens": prompt_tokens})
        return _json_response(self, 404, {"error": "Not Found"})

    def do_POST(self):
        parsed = urlparse(self.path)
        if parsed.path == "/shutdown":
            if not self._client_is_loopback():
                return _json_response(self, 403, {"error": "Forbidden"})

            _json_response(self, 200, {"status": "shutting_down"})

            def _shutdown():
                try:
                    self.server.shutdown()
                except Exception:
                    logger.exception("Shutdown request failed")
                try:
                    self.server.server_close()
                except Exception:
                    pass

            threading.Thread(target=_shutdown, daemon=True).start()
            self.close_connection = True
            return
        if parsed.path == "/v1/messages/count_tokens":
            try:
                length = int(self.headers.get("Content-Length", "0"))
                raw_body = self.rfile.read(length) if length else b"{}"
                incoming = json.loads(raw_body.decode("utf-8") or "{}")
            except Exception as exc:
                logger.exception("Failed to parse count_tokens request")
                return _json_response(self, 400, {"error": f"Invalid JSON: {exc}"})
            prompt_tokens = streaming.estimate_prompt_tokens(incoming)
            return _json_response(self, 200, {"input_tokens": prompt_tokens})
        if parsed.path != "/v1/messages":
            return _json_response(self, 404, {"error": "Not Found"})

        try:
            length = int(self.headers.get("Content-Length", "0"))
            raw_body = self.rfile.read(length) if length else b"{}"
            incoming = json.loads(raw_body.decode("utf-8") or "{}")
        except Exception as exc:
            logger.exception("Failed to parse incoming request")
            return _json_response(self, 400, {"error": f"Invalid JSON: {exc}"})

        log_payload(logger, "Incoming /v1/messages payload", incoming)

        try:
            requested_model = incoming.get("model")
            provider, target_model = resolve_provider_model(requested_model, self.settings)
            normalized_model = False
        except ValueError as exc:
            return _json_response(self, 400, {"error": str(exc)})

        effective_settings = self.settings
        codex_haiku_override = ""
        if provider == "codex":
            effective_settings, effective_model = _effective_codex_settings(
                self.settings, requested_model, target_model
            )
            if effective_model != target_model:
                codex_haiku_override = effective_model
                target_model = effective_model

        resolution_bits = []
        if requested_model and requested_model != target_model:
            resolution_bits.append(f"requested={requested_model}")
        if normalized_model:
            resolution_bits.append("alias->canonical")
        if (
            provider == "lmstudio"
            and requested_model
            and "claude-haiku" in requested_model.lower()
            and target_model == self.settings.lmstudio_model
        ):
            resolution_bits.append("haiku->lmstudio_default")
        if codex_haiku_override:
            resolution_bits.append(f"haiku->{codex_haiku_override}")
        suffix = f" ({'; '.join(resolution_bits)})" if resolution_bits else ""
        logger.info("Resolved model %s:%s%s", provider, target_model, suffix)

        try:
            openai_payload = anthropic_to_openai(incoming, target_model)
        except Exception as exc:
            logger.exception("Failed to translate Anthropic request")
            return _json_response(self, 400, {"error": f"Bad request: {exc}"})

        log_payload(
            logger,
            f"Sending payload to {provider}:{target_model}",
            openai_payload,
        )

        # Poe direct (OpenAI-compatible API)
        if provider == "poe":
            if not self.settings.poe_api_key:
                return _json_response(self, 400, {"error": "POE_API_KEY not set"})
            if openai_payload.get("stream"):
                return self._handle_poe_stream(openai_payload, target_model, incoming)
            return self._handle_poe(openai_payload, target_model, incoming)

        # OpenRouter
        if provider == "openrouter":
            if not self.settings.openrouter_key:
                return _json_response(self, 400, {"error": "OPENROUTER_API_KEY not set"})
            if openai_payload.get("stream"):
                return self._handle_openrouter_stream(
                    openai_payload, target_model, incoming
                )
            try:
                openrouter_response = openrouter.send(openai_payload, self.settings, target_model)
                outgoing = openai_to_anthropic(openrouter_response, target_model, incoming)
                log_payload(logger, "OpenRouter response", openrouter_response)
                log_payload(logger, "Responding to client", outgoing)
                return _json_response(self, 200, outgoing)
            except Exception as exc:
                logger.exception("OpenRouter request failed")
                return _json_response(self, 502, {"error": f"OpenRouter error: {exc}"})

        # OpenAI Codex (ChatGPT OAuth)
        if provider == "codex":
            if openai_payload.get("stream"):
                return self._handle_codex_stream(openai_payload, target_model, incoming, effective_settings)
            try:
                codex_response = codex.send(openai_payload, effective_settings, target_model)
                outgoing = openai_to_anthropic(codex_response, target_model, incoming)
                log_payload(logger, "Codex response", codex_response)
                log_payload(logger, "Responding to client", outgoing)
                return _json_response(self, 200, outgoing)
            except Exception as exc:
                logger.exception("Codex request failed")
                return _json_response(self, 502, {"error": f"Codex error: {exc}"})

        # LM Studio
        if openai_payload.get("stream"):
            return self._handle_lm_stream(openai_payload, target_model, incoming)

        try:
            lmstudio_response = lmstudio.send(openai_payload, self.settings)
            outgoing = openai_to_anthropic(lmstudio_response, target_model, incoming)
            log_payload(logger, "LM Studio response", lmstudio_response)
            log_payload(logger, "Responding to client", outgoing)
            return _json_response(self, 200, outgoing)
        except Exception as exc:
            logger.exception("LM Studio request failed")
            return _json_response(self, 502, {"error": f"LM Studio error: {exc}"})

    # Provider dispatch helpers
    def _handle_poe(self, payload: Dict[str, Any], bot_name: str, incoming: Dict[str, Any]):
        try:
            poe_response = poe.send(payload, self.settings, bot_name, incoming)
            log_payload(logger, "Responding to client", poe_response)
            return _json_response(self, 200, poe_response)
        except Exception as exc:
            logger.exception("Poe request failed")
            return _json_response(self, 502, {"error": f"Poe error: {exc}"})

    def _handle_poe_stream(
        self, payload: Dict[str, Any], bot_name: str, incoming: Dict[str, Any]
    ):
        try:
            return poe.stream(payload, self.settings, bot_name, incoming, self, logger)
        except (BrokenPipeError, ConnectionResetError):
            logger.info("Client disconnected during Poe stream")
            self.close_connection = True
            return
        except Exception as exc:
            logger.exception("Poe stream failed")
            return _json_response(self, 502, {"error": f"Poe error: {exc}"})

    def _handle_openrouter_stream(
        self, payload: Dict[str, Any], target_model: str, incoming: Dict[str, Any]
    ):
        try:
            return openrouter.stream(payload, self.settings, target_model, incoming, self, logger)
        except (BrokenPipeError, ConnectionResetError):
            logger.info("Client disconnected during OpenRouter stream")
            self.close_connection = True
            return
        except Exception as exc:
            logger.exception("OpenRouter stream failed")
            return _json_response(self, 502, {"error": f"OpenRouter error: {exc}"})

    def _handle_codex_stream(
        self,
        payload: Dict[str, Any],
        target_model: str,
        incoming: Dict[str, Any],
        settings: Optional[Settings] = None,
    ):
        try:
            return codex.stream(payload, settings or self.settings, target_model, incoming, self, logger)
        except (BrokenPipeError, ConnectionResetError):
            logger.info("Client disconnected during Codex stream")
            self.close_connection = True
            return
        except Exception as exc:
            logger.exception("Codex stream failed")
            return _json_response(self, 502, {"error": f"Codex error: {exc}"})

    def _handle_lm_stream(
        self, payload: Dict[str, Any], target_model: str, incoming: Dict[str, Any]
    ):
        try:
            return lmstudio.stream(payload, self.settings, target_model, incoming, self, logger)
        except (BrokenPipeError, ConnectionResetError):
            logger.info("Client disconnected during LM Studio stream")
            self.close_connection = True
            return
        except Exception as exc:
            logger.exception("LM Studio stream failed")
            return _json_response(self, 502, {"error": f"LM Studio error: {exc}"})


def run_server(settings: Settings):
    server = build_server(settings)
    logger.info(
        "Adapter listening on http://%s:%s (model=%s)",
        settings.host,
        settings.port,
        settings.model or "client-provided-only",
    )
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        logger.info("Shutting down adapter")
    finally:
        server.server_close()


def build_server(settings: Settings) -> AdapterHTTPServer:
    """Create a configured AdapterHTTPServer without starting it."""
    settings.apply_no_proxy_env()
    server = AdapterHTTPServer((settings.host, settings.port), AdapterHandler)
    AdapterHandler.settings = settings  # type: ignore
    return server


def _ensure_codex_login_if_needed(settings: Settings, no_browser: bool, allow_interactive: bool) -> None:
    model = (settings.model or "").strip()
    if not model.lower().startswith("codex:"):
        return

    auth_mode = str(getattr(settings, "codex_auth", "") or "").strip().lower()
    env_present = bool(
        getattr(settings, "codex_access_token", "")
        and getattr(settings, "codex_refresh_token", "")
        and int(getattr(settings, "codex_expires_at_ms", 0) or 0) > 0
    )

    from .codex_oauth import load_tokens

    file_present = load_tokens() is not None

    if auth_mode in {"env", "token", "tokens"}:
        if not env_present:
            raise RuntimeError(
                "Codex auth mode is env but tokens are missing (set OPENAI_CODEX_ACCESS_TOKEN/OPENAI_CODEX_REFRESH_TOKEN/OPENAI_CODEX_EXPIRES_AT_MS)."
            )
        return

    if auth_mode in {"oauth", "login", "file", "stored"}:
        if file_present:
            return
        if not allow_interactive:
            raise RuntimeError("OpenAI Codex OAuth not configured (run cc-adapter-codex-login).")
        logger.info("OpenAI Codex OAuth not found. Starting OAuth login flow...")
        from .codex_auth import login

        login(no_browser=no_browser)
        return

    # auto
    if env_present or file_present:
        return
    if not allow_interactive:
        raise RuntimeError("OpenAI Codex OAuth not configured (run cc-adapter-codex-login).")
    logger.info("OpenAI Codex OAuth not found. Starting OAuth login flow...")
    from .codex_auth import login

    login(no_browser=no_browser)


def main():
    parser = argparse.ArgumentParser(description="LM Studio / Poe / OpenRouter / OpenAI Codex adapter")
    parser.add_argument("--host", default=os.getenv("ADAPTER_HOST", "127.0.0.1"))
    parser.add_argument("--port", type=int, default=int(os.getenv("ADAPTER_PORT", "8005")))
    parser.add_argument("--daemon", action="store_true", help="run in background")
    parser.add_argument(
        "--model",
        required=False,
        help="Default model as provider:model (e.g., poe:claude-opus-4.5) or provider-only shorthand (e.g., codex:). Can also set CC_ADAPTER_MODEL.",
    )
    parser.add_argument(
        "--provider",
        required=False,
        help="Provider prefix to pair with an unprefixed --model (poe|lmstudio|openrouter|codex). Example: --provider codex --model gpt-5.2-high",
    )
    parser.add_argument(
        "--context-window",
        type=int,
        default=int(os.getenv("CONTEXT_WINDOW", "0")),
        help="Context window in tokens (prompt + completion); defaults per model if omitted.",
    )
    parser.add_argument("--lmstudio-base", help="LM Studio base URL (OpenAI compatible)")
    parser.add_argument("--lmstudio-model", help="LM Studio model name")
    parser.add_argument("--lmstudio-timeout", type=float, help="LM Studio timeout (seconds)")
    parser.add_argument("--poe-api-key", help="Poe API key")
    parser.add_argument("--poe-base-url", help="Poe base URL")
    parser.add_argument("--poe-max-retries", type=int, help="Poe retry attempts for upstream errors")
    parser.add_argument("--poe-retry-backoff", type=float, help="Seconds between retry attempts")
    parser.add_argument("--openrouter-api-key", help="OpenRouter API key")
    parser.add_argument("--openrouter-base", help="OpenRouter base URL")
    parser.add_argument("--codex-base-url", help="OpenAI Codex base URL (ChatGPT backend)")
    parser.add_argument(
        "--codex-auth",
        choices=["auto", "oauth", "env"],
        help="Codex auth mode: auto (env tokens or stored OAuth), oauth (stored OAuth only), env (env tokens only).",
    )
    parser.add_argument(
        "--codex-no-browser",
        action="store_true",
        help="Do not attempt to open a browser during Codex OAuth login (manual paste fallback).",
    )
    args = parser.parse_args()

    model_arg = args.model
    provider_arg = str(getattr(args, "provider", "") or "").strip()
    if provider_arg:
        provider = provider_arg.rstrip(":").lower()
        if provider not in {"poe", "lmstudio", "openrouter", "codex"}:
            parser.error(f"Unsupported provider: {provider_arg}")
        if model_arg:
            if ":" in model_arg:
                model_provider = model_arg.split(":", 1)[0].lower()
                if model_provider != provider:
                    parser.error(f"--provider {provider_arg} conflicts with --model {model_arg}")
            else:
                model_arg = f"{provider}:{model_arg}"
        else:
            model_arg = f"{provider}:"
    model_arg = normalize_model_spec(model_arg)

    overrides = {
        "host": args.host,
        "port": args.port,
        "model": model_arg,
        "context_window": args.context_window,
        "lmstudio_base": args.lmstudio_base,
        "lmstudio_model": args.lmstudio_model,
        "lmstudio_timeout": args.lmstudio_timeout,
        "poe_api_key": args.poe_api_key,
        "poe_base_url": args.poe_base_url,
        "poe_max_retries": args.poe_max_retries,
        "poe_retry_backoff": args.poe_retry_backoff,
        "openrouter_key": args.openrouter_api_key,
        "openrouter_base": args.openrouter_base,
        "codex_base_url": args.codex_base_url,
        "codex_auth": args.codex_auth,
    }

    if args.daemon:
        if not port_available(args.host, args.port):
            logger.error(
                "Port %s:%s already in use. Stop the existing adapter or choose another port.",
                args.host,
                args.port,
            )
            sys.exit(1)
        try:
            settings = apply_overrides(load_settings(), overrides)
            settings.model = normalize_model_spec(settings.model) or settings.model
            _ensure_codex_login_if_needed(
                settings,
                no_browser=bool(args.codex_no_browser),
                allow_interactive=sys.stdin.isatty(),
            )
        except Exception as exc:
            logger.error("Codex authentication failed: %s", exc)
            sys.exit(1)
        cmd = [sys.executable, "-m", "cc_adapter.server", "--host", args.host, "--port", str(args.port)]
        if model_arg:
            cmd.extend(["--model", model_arg])
        if args.context_window:
            cmd.extend(["--context-window", str(args.context_window)])
        if args.lmstudio_base:
            cmd.extend(["--lmstudio-base", args.lmstudio_base])
        if args.lmstudio_model:
            cmd.extend(["--lmstudio-model", args.lmstudio_model])
        if args.lmstudio_timeout:
            cmd.extend(["--lmstudio-timeout", str(args.lmstudio_timeout)])
        if args.poe_api_key:
            cmd.extend(["--poe-api-key", args.poe_api_key])
        if args.poe_base_url:
            cmd.extend(["--poe-base-url", args.poe_base_url])
        if args.poe_max_retries is not None:
            cmd.extend(["--poe-max-retries", str(args.poe_max_retries)])
        if args.poe_retry_backoff is not None:
            cmd.extend(["--poe-retry-backoff", str(args.poe_retry_backoff)])
        if args.openrouter_api_key:
            cmd.extend(["--openrouter-api-key", args.openrouter_api_key])
        if args.openrouter_base:
            cmd.extend(["--openrouter-base", args.openrouter_base])
        if args.codex_base_url:
            cmd.extend(["--codex-base-url", args.codex_base_url])
        if args.codex_auth:
            cmd.extend(["--codex-auth", args.codex_auth])
        proc = Popen(cmd, stdout=DEVNULL, stderr=DEVNULL, stdin=DEVNULL, close_fds=True)
        print(f"Started daemon pid={proc.pid}")
        return

    if not port_available(args.host, args.port):
        logger.error(
            "Port %s:%s already in use. Stop the existing adapter or choose another port.",
            args.host,
            args.port,
        )
        sys.exit(1)

    settings = apply_overrides(load_settings(), overrides)
    settings.model = normalize_model_spec(settings.model) or settings.model
    try:
        _ensure_codex_login_if_needed(
            settings,
            no_browser=bool(args.codex_no_browser),
            allow_interactive=sys.stdin.isatty(),
        )
    except Exception as exc:
        logger.error("Codex authentication failed: %s", exc)
        sys.exit(1)
    try:
        run_server(settings)
    except OSError:
        logger.exception("Failed to start server")
        sys.exit(1)


if __name__ == "__main__":
    main()



import json
import logging
import os
import threading
import time
from pathlib import Path
from typing import Dict, Literal, Optional, Tuple

import platformdirs
import requests

logger = logging.getLogger(__name__)

ModelFamily = Literal["codex-max", "codex", "gpt-5.2", "gpt-5.1"]

GITHUB_API_RELEASES = "https://api.github.com/repos/openai/codex/releases/latest"
GITHUB_HTML_RELEASES = "https://github.com/openai/codex/releases/latest"

# These map to Codex CLI prompt selection (see ref/opencode-openai-codex-auth).
PROMPT_FILES: Dict[ModelFamily, str] = {
    "codex-max": "gpt-5.1-codex-max_prompt.md",
    "codex": "gpt_5_codex_prompt.md",
    "gpt-5.2": "gpt-5.1-codex-max_prompt.md",  # GPT-5.2 uses the same prompt as codex-max.
    "gpt-5.1": "gpt_5_1_prompt.md",
}

CACHE_FILES: Dict[ModelFamily, str] = {
    "codex-max": "codex-max-instructions.md",
    "codex": "codex-instructions.md",
    "gpt-5.2": "gpt-5.2-instructions.md",
    "gpt-5.1": "gpt-5.1-instructions.md",
}

_CACHE_TTL_MS = 15 * 60 * 1000  # 15 minutes
_LOCK = threading.Lock()
_MEMO: Dict[ModelFamily, Tuple[int, str]] = {}


def _config_dir() -> Path:
    override = os.getenv("CC_ADAPTER_CONFIG_DIR", "").strip()
    if override:
        return Path(override)
    return Path(platformdirs.user_config_dir("cc-adapter"))


def _cache_dir() -> Path:
    return _config_dir() / "cache"


def _cache_paths(model_family: ModelFamily) -> Tuple[Path, Path]:
    cache_file = _cache_dir() / CACHE_FILES[model_family]
    meta_file = cache_file.with_name(cache_file.name.replace(".md", "-meta.json"))
    return cache_file, meta_file


def model_family_for(model: str) -> ModelFamily:
    normalized = (model or "").strip().lower()
    if "codex-max" in normalized:
        return "codex-max"
    if "codex" in normalized or normalized.startswith("codex-"):
        return "codex"
    if "gpt-5.2" in normalized:
        return "gpt-5.2"
    return "gpt-5.1"


def _load_meta(meta_file: Path) -> Dict[str, object]:
    try:
        if meta_file.is_file():
            return json.loads(meta_file.read_text(encoding="utf-8"))
    except Exception:
        return {}
    return {}


def _write_meta(meta_file: Path, payload: Dict[str, object]) -> None:
    meta_file.parent.mkdir(parents=True, exist_ok=True)
    tmp = meta_file.with_suffix(meta_file.suffix + ".tmp")
    tmp.write_text(json.dumps(payload, ensure_ascii=False, indent=2), encoding="utf-8")
    tmp.replace(meta_file)


def _latest_release_tag(
    proxies: Optional[Dict[str, str]] = None,
    timeout: float = 30.0,
) -> str:
    try:
        resp = requests.get(GITHUB_API_RELEASES, timeout=timeout, proxies=proxies)
        if resp.ok:
            data = resp.json() if resp.content else {}
            tag = data.get("tag_name")
            if tag:
                return str(tag)
    except Exception:
        pass

    resp = requests.get(GITHUB_HTML_RELEASES, timeout=timeout, proxies=proxies, allow_redirects=True)
    resp.raise_for_status()

    final_url = str(resp.url or "")
    if "/tag/" in final_url:
        tag = final_url.split("/tag/", 1)[1].split("/", 1)[0].strip()
        if tag:
            return tag

    html = resp.text or ""
    marker = "/openai/codex/releases/tag/"
    idx = html.find(marker)
    if idx != -1:
        start = idx + len(marker)
        end = html.find('"', start)
        if end != -1:
            tag = html[start:end].strip()
            if tag and "/" not in tag:
                return tag

    raise RuntimeError("Failed to determine latest openai/codex release tag")


def _raw_prompt_url(tag: str, prompt_file: str) -> str:
    return f"https://raw.githubusercontent.com/openai/codex/{tag}/codex-rs/core/{prompt_file}"


def get_codex_instructions(
    model: str,
    *,
    proxies: Optional[Dict[str, str]] = None,
    timeout: float = 30.0,
    force_refresh: bool = False,
) -> str:
    """
    Return the exact Codex CLI system prompt required by the ChatGPT Codex backend.

    The ChatGPT Codex backend rejects arbitrary `instructions` payloads with:
      {"detail":"Instructions are not valid"}
    """

    family = model_family_for(model)
    cache_file, meta_file = _cache_paths(family)
    now_ms = int(time.time() * 1000)

    with _LOCK:
        if not force_refresh and family in _MEMO:
            cached_at_ms, cached = _MEMO[family]
            if cached and now_ms - cached_at_ms < _CACHE_TTL_MS:
                return cached

    meta = _load_meta(meta_file)
    cached_etag = str(meta.get("etag") or "").strip() or None
    cached_tag = str(meta.get("tag") or "").strip() or None
    last_checked_ms = int(meta.get("last_checked_ms") or 0)

    if (
        not force_refresh
        and cache_file.is_file()
        and last_checked_ms > 0
        and now_ms - last_checked_ms < _CACHE_TTL_MS
    ):
        text = cache_file.read_text(encoding="utf-8")
        with _LOCK:
            _MEMO[family] = (now_ms, text)
        return text

    try:
        latest_tag = _latest_release_tag(proxies=proxies, timeout=timeout)
        prompt_file = PROMPT_FILES[family]
        url = _raw_prompt_url(latest_tag, prompt_file)

        etag = cached_etag
        if force_refresh or (cached_tag and cached_tag != latest_tag):
            etag = None

        headers: Dict[str, str] = {}
        if etag and not force_refresh:
            headers["If-None-Match"] = etag

        resp = requests.get(url, headers=headers, timeout=timeout, proxies=proxies)
        if resp.status_code == 304 and cache_file.is_file() and not force_refresh:
            text = cache_file.read_text(encoding="utf-8")
            _write_meta(
                meta_file,
                {
                    "etag": cached_etag,
                    "tag": latest_tag,
                    "last_checked_ms": now_ms,
                    "url": url,
                },
            )
            with _LOCK:
                _MEMO[family] = (now_ms, text)
            return text

        resp.raise_for_status()
        text = resp.text or ""
        if not text.strip():
            raise RuntimeError("Fetched Codex instructions were empty")

        cache_file.parent.mkdir(parents=True, exist_ok=True)
        cache_file.write_text(text, encoding="utf-8")
        _write_meta(
            meta_file,
            {
                "etag": resp.headers.get("etag") or resp.headers.get("ETag"),
                "tag": latest_tag,
                "last_checked_ms": now_ms,
                "url": url,
            },
        )
        with _LOCK:
            _MEMO[family] = (now_ms, text)
        return text
    except Exception as exc:
        if cache_file.is_file():
            logger.warning(
                "Failed to refresh Codex instructions (%s); using cached version: %s",
                family,
                exc,
            )
            text = cache_file.read_text(encoding="utf-8")
            with _LOCK:
                _MEMO[family] = (now_ms, text)
            return text
        raise RuntimeError(f"Failed to fetch Codex instructions ({family}): {exc}") from exc




import argparse
import sys
import webbrowser

from .codex_oauth import (
    build_authorization_url,
    create_state,
    delete_tokens,
    exchange_authorization_code,
    extract_chatgpt_account_id,
    generate_pkce_pair,
    parse_authorization_input,
    save_tokens,
    start_local_callback_server,
    wait_for_callback_code,
)


def _prompt_manual_code(expected_state: str) -> str:
    print("Paste the full callback URL or the authorization code (code#state also supported):")
    raw = sys.stdin.readline().strip()
    code, state = parse_authorization_input(raw)
    if not code:
        raise SystemExit("No authorization code provided")
    if state and state != expected_state:
        raise SystemExit("State mismatch")
    return code


def login(no_browser: bool = False) -> int:
    verifier, challenge = generate_pkce_pair()
    state = create_state()
    url = build_authorization_url(state, challenge)

    print("OpenAI Codex OAuth login (ChatGPT subscription)")
    print(f"Auth URL:\n{url}\n")

    server = None
    code = None
    try:
        server = start_local_callback_server(state)
        if not no_browser:
            webbrowser.open(url, new=1, autoraise=True)
        code = wait_for_callback_code(server, timeout_seconds=300)
    except OSError:
        server = None
    finally:
        if server:
            try:
                server.shutdown()
            except Exception:
                pass
            try:
                server.server_close()
            except Exception:
                pass

    if not code:
        if not no_browser:
            webbrowser.open(url, new=1, autoraise=True)
        code = _prompt_manual_code(state)

    tokens = exchange_authorization_code(code=code, code_verifier=verifier)
    token_path = save_tokens(tokens)
    account_id = extract_chatgpt_account_id(tokens.access)

    print(f"Saved OAuth tokens to: {token_path}")
    if account_id:
        print(f"ChatGPT account id: {account_id}")
    else:
        print("Warning: failed to extract ChatGPT account id from access token.")
    return 0


def main() -> None:
    parser = argparse.ArgumentParser(description="Authenticate cc-adapter for OpenAI Codex (ChatGPT OAuth)")
    parser.add_argument("--no-browser", action="store_true", help="Do not attempt to open a browser automatically")
    parser.add_argument("--logout", action="store_true", help="Delete the stored OAuth token file")
    args = parser.parse_args()

    if args.logout:
        delete_tokens()
        print("Deleted stored OpenAI Codex OAuth tokens.")
        raise SystemExit(0)

    raise SystemExit(login(no_browser=bool(args.no_browser)))


if __name__ == "__main__":
    main()



import unittest

import base64
import json
import os
import tempfile
from pathlib import Path
from unittest import mock

from cc_adapter.providers import codex


def _jwt(payload: dict) -> str:
    header = {"alg": "none", "typ": "JWT"}

    def b64url(obj: dict) -> str:
        return base64.urlsafe_b64encode(json.dumps(obj).encode("utf-8")).decode("utf-8").rstrip("=")

    return f"{b64url(header)}.{b64url(payload)}.sig"


class CodexTransformTestCase(unittest.TestCase):
    def test_messages_to_responses_input_uses_output_text_for_assistant(self):
        messages = [
            {"role": "user", "content": "hi"},
            {"role": "assistant", "content": [{"type": "text", "text": "hello"}]},
        ]

        _, items = codex._messages_to_responses_input(messages)
        self.assertEqual(items[0]["type"], "message")
        self.assertEqual(items[0]["role"], "user")
        self.assertEqual(items[0]["content"][0]["type"], "input_text")
        self.assertEqual(items[0]["content"][0]["text"], "hi")

        self.assertEqual(items[1]["type"], "message")
        self.assertEqual(items[1]["role"], "assistant")
        self.assertEqual(items[1]["content"][0]["type"], "output_text")
        self.assertEqual(items[1]["content"][0]["text"], "hello")

    def test_messages_to_responses_input_maps_tools(self):
        messages = [
            {"role": "system", "content": "sys"},
            {"role": "user", "content": [{"type": "text", "text": "hi"}]},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {"name": "do_it", "arguments": "{\"x\":1}"},
                    }
                ],
            },
            {"role": "tool", "tool_call_id": "call_1", "content": "done"},
        ]

        developer_prompt, items = codex._messages_to_responses_input(messages)
        self.assertIn("sys", developer_prompt)

        function_calls = [i for i in items if i.get("type") == "function_call"]
        self.assertEqual(len(function_calls), 1)
        self.assertEqual(function_calls[0]["call_id"], "call_1")
        self.assertEqual(function_calls[0]["name"], "do_it")
        self.assertEqual(function_calls[0]["arguments"], "{\"x\":1}")

        outputs = [i for i in items if i.get("type") == "function_call_output"]
        self.assertEqual(len(outputs), 1)
        self.assertEqual(outputs[0]["call_id"], "call_1")
        self.assertEqual(outputs[0]["output"], "done")

    def test_responses_tools_flattens_chat_tools(self):
        tools = [
            {
                "type": "function",
                "function": {
                    "name": "web_search",
                    "description": "search",
                    "parameters": {"type": "object", "properties": {}},
                },
            }
        ]

        out = codex._responses_tools(tools)
        self.assertIsInstance(out, list)
        self.assertEqual(out[0]["type"], "function")
        self.assertEqual(out[0]["name"], "web_search")
        self.assertIn("parameters", out[0])

    def test_resolve_codex_auth_env_mode_uses_env_tokens(self):
        from cc_adapter.config import Settings
        from cc_adapter.codex_oauth import JWT_ACCOUNT_ID_FIELD, JWT_CLAIM_PATH

        token = _jwt({JWT_CLAIM_PATH: {JWT_ACCOUNT_ID_FIELD: "account-123"}})
        settings = Settings(
            model="codex:gpt-5.1-codex",
            codex_auth="env",
            codex_access_token=token,
            codex_refresh_token="refresh",
            codex_expires_at_ms=9_999_999_999_999,
            lmstudio_timeout=1,
        )

        tokens, account_id = codex._resolve_codex_auth(settings)
        self.assertEqual(account_id, "account-123")
        self.assertEqual(tokens.access, token)

    def test_resolve_codex_auth_oauth_mode_uses_token_file(self):
        from cc_adapter.config import Settings
        from cc_adapter.codex_oauth import DEFAULT_TOKEN_FILENAME, JWT_ACCOUNT_ID_FIELD, JWT_CLAIM_PATH

        token = _jwt({JWT_CLAIM_PATH: {JWT_ACCOUNT_ID_FIELD: "account-456"}})
        with tempfile.TemporaryDirectory() as cfgdir:
            with mock.patch.dict(os.environ, {"CC_ADAPTER_CONFIG_DIR": cfgdir}, clear=False):
                Path(cfgdir, DEFAULT_TOKEN_FILENAME).write_text(
                    json.dumps(
                        {
                            "type": "oauth",
                            "access": token,
                            "refresh": "refresh",
                            "expires_at_ms": 9_999_999_999_999,
                        }
                    ),
                    encoding="utf-8",
                )
                settings = Settings(model="codex:gpt-5.1-codex", codex_auth="oauth", lmstudio_timeout=1)

                tokens, account_id = codex._resolve_codex_auth(settings)
                self.assertEqual(account_id, "account-456")
                self.assertEqual(tokens.access, token)

    def test_request_body_applies_preset_defaults(self):
        from cc_adapter.config import Settings

        payload = {
            "model": "gpt-5.2",
            "messages": [{"role": "user", "content": "hi"}],
            "stream": True,
        }
        settings = Settings(lmstudio_timeout=1)
        with mock.patch.object(codex, "get_codex_instructions", return_value="codex-cli-prompt"):
            body = codex._request_body(payload, settings, model_key="codex:gpt-5.2-high")
        self.assertEqual(body["model"], "gpt-5.2")
        self.assertEqual(body["instructions"], "codex-cli-prompt")
        self.assertEqual(body["reasoning"]["effort"], "high")
        self.assertEqual(body["reasoning"]["summary"], "detailed")
        self.assertEqual(body["text"]["verbosity"], "medium")
        self.assertIs(body["store"], False)
        self.assertIn("reasoning.encrypted_content", body.get("include", []))

    def test_request_body_includes_reasoning_summary_and_text_defaults(self):
        from cc_adapter.config import Settings

        payload = {
            "model": "gpt-5.1-codex",
            "messages": [{"role": "user", "content": "hi"}],
            "stream": True,
        }
        settings = Settings(lmstudio_timeout=1)
        with mock.patch.object(codex, "get_codex_instructions", return_value="codex-cli-prompt"):
            body = codex._request_body(payload, settings, model_key="codex:gpt-5.1-codex")
        self.assertEqual(body["reasoning"]["effort"], "medium")
        self.assertEqual(body["reasoning"]["summary"], "auto")
        self.assertEqual(body["text"]["verbosity"], "medium")

    def test_request_body_drops_temperature_and_top_p(self):
        from cc_adapter.config import Settings

        payload = {
            "model": "gpt-5.2",
            "messages": [{"role": "user", "content": "hi"}],
            "stream": True,
            "temperature": 0.2,
            "top_p": 0.9,
        }
        settings = Settings(lmstudio_timeout=1)
        with mock.patch.object(codex, "get_codex_instructions", return_value="codex-cli-prompt"):
            body = codex._request_body(payload, settings, model_key="codex:gpt-5.2-medium")
        self.assertNotIn("temperature", body)
        self.assertNotIn("top_p", body)

    def test_request_body_preset_overrides_payload_reasoning(self):
        from cc_adapter.config import Settings

        payload = {
            "model": "gpt-5.2",
            "messages": [{"role": "user", "content": "hi"}],
            "stream": True,
            "reasoning": {"effort": "high", "summary": "detailed"},
            "text": {"verbosity": "high"},
        }
        settings = Settings(lmstudio_timeout=1)
        with mock.patch.object(codex, "get_codex_instructions", return_value="codex-cli-prompt"):
            body = codex._request_body(payload, settings, model_key="codex:gpt-5.2-low")
        self.assertEqual(body["reasoning"]["effort"], "low")
        self.assertEqual(body["reasoning"]["summary"], "auto")
        self.assertEqual(body["text"]["verbosity"], "medium")


if __name__ == "__main__":
    unittest.main()



import http.client
import json
import os
import threading
from pathlib import Path
from typing import List, Tuple
from dataclasses import replace

import pytest

from cc_adapter import server
from cc_adapter.config import Settings
from cc_adapter.model_registry import provider_model_slugs

# Match GUI dropdown models (provider-prefixed here).
GUI_TEST_MODELS = [f"poe:{name}" for name in provider_model_slugs("poe")] + [
    f"openrouter:{name}" for name in provider_model_slugs("openrouter")
]


def _load_env_file(env_path: Path) -> None:
    if not env_path.is_file():
        return
    for line in env_path.read_text().splitlines():
        line = line.strip()
        if not line or line.startswith("#") or "=" not in line:
            continue
        key, value = line.split("=", 1)
        key = key.strip()
        value = value.strip().strip("'").strip('"')
        os.environ.setdefault(key, value)


def _filter_models_by_keys(models: List[str], poe_key: str, openrouter_key: str) -> List[str]:
    filtered: List[str] = []
    for model in models:
        if model.startswith("poe:") and not poe_key:
            continue
        if model.startswith("openrouter:") and not openrouter_key:
            continue
        filtered.append(model)
    return filtered


def _start_server(settings: Settings):
    http_server = server.build_server(settings)
    thread = threading.Thread(target=http_server.serve_forever, daemon=True)
    thread.start()
    return http_server, thread


def _stop_server(http_server, thread):
    http_server.shutdown()
    http_server.server_close()
    thread.join(timeout=5)


@pytest.fixture(scope="session")
def live_settings() -> Tuple[Settings, List[str]]:
    if os.environ.get("RUN_LIVE_PROVIDER_TESTS") != "1":
        pytest.skip("Set RUN_LIVE_PROVIDER_TESTS=1 to run live provider connectivity tests")
    env_path = Path(__file__).resolve().parent.parent / ".env"
    _load_env_file(env_path)
    poe_key = (os.environ.get("POE_API_KEY") or "").strip()
    openrouter_key = (os.environ.get("OPENROUTER_API_KEY") or "").strip()
    if not poe_key and not openrouter_key:
        pytest.skip("No provider API keys found in environment/.env (POE_API_KEY/OPENROUTER_API_KEY)")
    models = _filter_models_by_keys(GUI_TEST_MODELS, poe_key, openrouter_key)
    settings = Settings(host="127.0.0.1", port=0, model=models[0])
    settings.poe_api_key = poe_key
    settings.openrouter_key = openrouter_key
    return settings, models


def test_gui_models_connectivity(live_settings, capsys):
    base_settings, models = live_settings
    for model in models:
        settings = replace(base_settings, model=model, port=0)
        http_server, thread = _start_server(settings)
        host, port = http_server.server_address
        try:
            conn = http.client.HTTPConnection(host, port, timeout=60)
            payload = {
                "model": model,
                "messages": [{"role": "user", "content": "Who are you? (respond in English)"}],
                "max_tokens": 16,
            }
            body = json.dumps(payload).encode("utf-8")
            conn.request("POST", "/v1/messages", body=body, headers={"Content-Type": "application/json"})
            resp = conn.getresponse()
            data = resp.read().decode("utf-8")
            conn.close()

            snippet = " ".join(data.split())
            with capsys.disabled():
                print(f"[live-test] {model} -> HTTP {resp.status} | {snippet}")

            if resp.status == 400 and "not a valid model" in data.lower():
                pytest.skip(f"{model} not accepted by provider: {snippet}")

            assert resp.status == 200, f"{model} -> {snippet}"
            parsed = json.loads(data)
            assert "error" not in parsed, f"{model} -> {snippet}"
            content = parsed.get("content") or []
            assert content, f"No content returned for {model}: {parsed}"
        finally:
            _stop_server(http_server, thread)



from typing import Optional, Tuple

from .config import Settings
from .codex_oauth import default_token_path
from .model_registry import DEFAULT_PROVIDER_MODELS, canonicalize_model, find_model, provider_models


def normalize_model_spec(model: Optional[str]) -> Optional[str]:
    """
    Normalize a CLI/env model spec.

    Supports provider-only shorthands:
    - "codex" or "codex:" -> "codex:<default>"
    - "poe:" -> "poe:<default>"
    """
    if model is None:
        return None
    raw = str(model).strip()
    if not raw:
        return raw

    normalized = raw.rstrip()
    if normalized.endswith(":"):
        provider = normalized[:-1].strip().lower()
        default = DEFAULT_PROVIDER_MODELS.get(provider)
        return f"{provider}:{default}" if default else raw

    if ":" not in normalized and normalized.lower() in DEFAULT_PROVIDER_MODELS:
        provider = normalized.lower()
        return f"{provider}:{DEFAULT_PROVIDER_MODELS[provider]}"

    if ":" in normalized:
        provider, name = normalized.split(":", 1)
        if not name.strip():
            default = DEFAULT_PROVIDER_MODELS.get(provider.lower())
            return f"{provider.lower()}:{default}" if default else raw

    return raw


def _display_slug(provider: str, name: str) -> str:
    match = find_model(provider, name)
    return match.slug if match else name


def available_models(settings: Settings) -> list[str]:
    """
    Build a user-facing list of provider-prefixed models based on available credentials.
    Ordering prefers higher-capability tiers first, while keeping any configured default at the top.
    """
    models: list[str] = []
    seen: set[str] = set()

    def add(provider: str, slug: str):
        key = f"{provider}:{slug}"
        if key not in seen:
            models.append(key)
            seen.add(key)

    # Always bubble the configured default to the top if it includes a prefix.
    if settings.model and ":" in settings.model:
        provider, name = settings.model.split(":", 1)
        provider = provider.lower()
        add(provider, _display_slug(provider, name))

    # Always show the LM Studio default so local-only users see an option.
    add("lmstudio", _display_slug("lmstudio", settings.lmstudio_model))

    # Include known provider offerings gated by available credentials.
    eligible: list[Tuple[int, str, str]] = []
    codex_env_present = bool(
        settings.codex_access_token
        and settings.codex_refresh_token
        and settings.codex_expires_at_ms > 0
    )
    codex_file_present = default_token_path().exists()
    codex_auth = str(getattr(settings, "codex_auth", "") or "").strip().lower()
    if codex_auth in {"env", "token", "tokens"}:
        codex_tokens_present = codex_env_present
    elif codex_auth in {"oauth", "login", "file", "stored"}:
        codex_tokens_present = codex_file_present
    else:
        codex_tokens_present = codex_env_present or codex_file_present
    if codex_tokens_present:
        for info in provider_models("codex"):
            eligible.append((info.priority, info.provider, info.slug))
    if settings.poe_api_key:
        for info in provider_models("poe"):
            eligible.append((info.priority, info.provider, info.slug))
    if settings.openrouter_key:
        for info in provider_models("openrouter"):
            eligible.append((info.priority, info.provider, info.slug))

    for _, provider, slug in sorted(eligible, key=lambda item: (item[0], item[1], item[2])):
        add(provider, slug)

    return models


def resolve_provider_model(model: Optional[str], settings: Settings) -> Tuple[str, str]:
    """Return (provider, upstream_model) while favoring the selected model by default."""
    default_model = settings.model
    if not default_model:
        raise ValueError("Model is required and must include provider prefix (e.g., poe:claude-opus-4.5)")
    if ":" not in default_model:
        raise ValueError("Model must include provider prefix (e.g., poe:claude-opus-4.5)")

    default_provider, default_name = default_model.split(":", 1)
    default_provider = default_provider.lower()
    if default_provider not in {"poe", "lmstudio", "openrouter", "codex"}:
        raise ValueError(f"Unsupported provider prefix: {default_provider}")

    requested = (model or "").strip()
    if requested:
        req_provider = default_provider
        req_name = requested
        if ":" in requested:
            req_provider, req_name = requested.split(":", 1)
            req_provider = req_provider.lower()

        if req_provider in {"poe", "openrouter"} and "claude-haiku" in req_name.lower():
            return req_provider, canonicalize_model(req_provider, req_name)

    return default_provider, canonicalize_model(default_provider, default_name)



import logging
import unittest

from cc_adapter import streaming


class DummyResponse:
    def __init__(self, lines):
        self._lines = list(lines)
        self.headers = {}
        self.closed = False

    def iter_lines(self, decode_unicode=False):
        yield from self._lines

    def close(self):
        self.closed = True


class DummyHandler:
    def __init__(self):
        self.buffer = b""
        self.close_connection = False
        self.wfile = self

    def write(self, data: bytes):
        self.buffer += data

    def flush(self):
        return


class ResponsesStreamingBridgeTestCase(unittest.TestCase):
    def test_streaming_text_delta(self):
        handler = DummyHandler()
        resp = DummyResponse(
            [
                b'data: {"type":"response.created","response_id":"resp_1"}',
                b'data: {"type":"response.output_text.delta","item_id":"msg_1","output_index":0,"content_index":0,"delta":"hello"}',
                b'data: {"type":"response.completed","response":{"id":"resp_1","usage":{"input_tokens":2,"output_tokens":1}}}',
            ]
        )
        logger = logging.getLogger("responses-stream-test")
        logger.setLevel(logging.CRITICAL)

        streaming.stream_responses_response(resp, "codex:gpt-5.1-codex", {}, handler, logger)

        body = handler.buffer.decode("utf-8")
        self.assertIn("event: message_start", body)
        self.assertIn("text_delta", body)
        self.assertIn("hello", body)
        self.assertIn("\"stop_reason\": \"end_turn\"", body)
        self.assertIn("\"input_tokens\": 2", body)

    def test_streaming_tool_call_arguments(self):
        handler = DummyHandler()
        resp = DummyResponse(
            [
                b'data: {"type":"response.created","response_id":"resp_2"}',
                b'data: {"type":"response.output_item.added","response_id":"resp_2","output_index":0,"item":{"type":"function_call","id":"fc_1","call_id":"call_1","name":"do_it","arguments":""}}',
                b'data: {"type":"response.function_call_arguments.delta","response_id":"resp_2","item_id":"fc_1","output_index":0,"delta":"{\\\"x\\\":1"}',
                b'data: {"type":"response.function_call_arguments.delta","response_id":"resp_2","item_id":"fc_1","output_index":0,"delta":"}"}',
                b'data: {"type":"response.completed","response":{"id":"resp_2","usage":{"input_tokens":3,"output_tokens":1}}}',
            ]
        )
        logger = logging.getLogger("responses-tool-test")
        logger.setLevel(logging.CRITICAL)

        streaming.stream_responses_response(resp, "codex:gpt-5.1-codex", {}, handler, logger)

        body = handler.buffer.decode("utf-8")
        self.assertIn("\"type\": \"tool_use\"", body)
        self.assertIn("\"id\": \"call_1\"", body)
        self.assertIn("\"name\": \"do_it\"", body)
        self.assertIn("input_json_delta", body)
        self.assertIn("\"stop_reason\": \"tool_use\"", body)


if __name__ == "__main__":
    unittest.main()




import logging
import unittest

from cc_adapter import streaming


class DummyResponse:
    def __init__(self):
        self.headers = {}
        self.closed = False

    def iter_lines(self, decode_unicode=False):
        yield b'data: {"choices": [{"delta": {"content": "hello"}}]}'
        raise ValueError("boom")

    def close(self):
        self.closed = True


class DummyHandler:
    def __init__(self):
        self.buffer = b""
        self.close_connection = False
        self.wfile = self

    def write(self, data: bytes):
        self.buffer += data

    def flush(self):
        return


class StreamingErrorHandlingTestCase(unittest.TestCase):
    def test_streaming_error_sends_stop_and_error_event(self):
        handler = DummyHandler()
        resp = DummyResponse()
        logger = logging.getLogger("stream-test")
        logger.setLevel(logging.CRITICAL)

        streaming.stream_openai_response(resp, "poe:deepseek-v3.2", {}, handler, logger)

        body = handler.buffer.decode("utf-8")
        self.assertTrue(handler.close_connection)
        self.assertIn("stop_reason\": \"error\"", body)
        self.assertIn("event: error", body)


if __name__ == "__main__":
    unittest.main()



import base64
import hashlib
import json
import os
import secrets
import threading
import time
from dataclasses import dataclass
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path
from typing import Any, Dict, Optional, Tuple
from urllib.parse import parse_qs, urlencode, urlparse

import platformdirs
import requests

CLIENT_ID = "app_EMoamEEZ73f0CkXaXp7hrann"
AUTHORIZE_URL = "https://auth.openai.com/oauth/authorize"
TOKEN_URL = "https://auth.openai.com/oauth/token"
REDIRECT_URI = "http://localhost:1455/auth/callback"
SCOPE = "openid profile email offline_access"

JWT_CLAIM_PATH = "https://api.openai.com/auth"
JWT_ACCOUNT_ID_FIELD = "chatgpt_account_id"

DEFAULT_TOKEN_FILENAME = "openai_codex_oauth.json"
OAUTH_CALLBACK_PORT = 1455

SUCCESS_HTML = """<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1" />
  <title>CC Adapter - OAuth Success</title>
  <style>
    body { font-family: -apple-system, BlinkMacSystemFont, Segoe UI, Roboto, Helvetica, Arial, sans-serif; padding: 24px; }
    code { background: #f3f4f6; padding: 2px 6px; border-radius: 6px; }
  </style>
</head>
<body>
  <h2>OAuth complete</h2>
  <p>You can close this tab and return to <code>cc-adapter</code>.</p>
</body>
</html>
"""


@dataclass(frozen=True)
class CodexOAuthTokens:
    access: str
    refresh: str
    expires_at_ms: int

    def expired(self, skew_seconds: int = 60) -> bool:
        return self.expires_at_ms <= int(time.time() * 1000) + (skew_seconds * 1000)

    def to_json(self) -> str:
        return json.dumps(
            {
                "type": "oauth",
                "access": self.access,
                "refresh": self.refresh,
                "expires_at_ms": self.expires_at_ms,
            },
            indent=2,
        )

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "CodexOAuthTokens":
        access = str(data.get("access") or "")
        refresh = str(data.get("refresh") or "")
        expires_at_ms = int(data.get("expires_at_ms") or 0)
        if not access or not refresh or expires_at_ms <= 0:
            raise ValueError("Invalid token payload")
        return CodexOAuthTokens(access=access, refresh=refresh, expires_at_ms=expires_at_ms)


def default_token_path() -> Path:
    override = os.getenv("CC_ADAPTER_CONFIG_DIR", "").strip()
    if override:
        return Path(override) / DEFAULT_TOKEN_FILENAME
    return Path(platformdirs.user_config_dir("cc-adapter")) / DEFAULT_TOKEN_FILENAME


def load_tokens(path: Optional[Path] = None) -> Optional[CodexOAuthTokens]:
    token_path = path or default_token_path()
    if not token_path.exists():
        return None
    try:
        raw = json.loads(token_path.read_text(encoding="utf-8"))
        return CodexOAuthTokens.from_dict(raw)
    except Exception:
        return None


def save_tokens(tokens: CodexOAuthTokens, path: Optional[Path] = None) -> Path:
    token_path = path or default_token_path()
    token_path.parent.mkdir(parents=True, exist_ok=True)
    tmp_path = token_path.with_suffix(token_path.suffix + ".tmp")
    tmp_path.write_text(tokens.to_json(), encoding="utf-8")
    try:
        os.chmod(tmp_path, 0o600)
    except Exception:
        pass
    tmp_path.replace(token_path)
    return token_path


def delete_tokens(path: Optional[Path] = None) -> None:
    token_path = path or default_token_path()
    try:
        token_path.unlink()
    except FileNotFoundError:
        return


def create_state() -> str:
    return secrets.token_hex(16)


def _b64url_no_pad(data: bytes) -> str:
    return base64.urlsafe_b64encode(data).decode("utf-8").rstrip("=")


def generate_pkce_pair() -> Tuple[str, str]:
    verifier = secrets.token_urlsafe(64)
    if len(verifier) > 128:
        verifier = verifier[:128]
    digest = hashlib.sha256(verifier.encode("utf-8")).digest()
    challenge = _b64url_no_pad(digest)
    return verifier, challenge


def build_authorization_url(state: str, code_challenge: str) -> str:
    params = {
        "response_type": "code",
        "client_id": CLIENT_ID,
        "redirect_uri": REDIRECT_URI,
        "scope": SCOPE,
        "code_challenge": code_challenge,
        "code_challenge_method": "S256",
        "state": state,
        "id_token_add_organizations": "true",
        "codex_cli_simplified_flow": "true",
        "originator": "codex_cli_rs",
    }
    return f"{AUTHORIZE_URL}?{urlencode(params)}"


def parse_authorization_input(value: str) -> Tuple[Optional[str], Optional[str]]:
    raw = (value or "").strip()
    if not raw:
        return None, None

    try:
        parsed = urlparse(raw)
        qs = parse_qs(parsed.query or "")
        code = (qs.get("code") or [None])[0]
        state = (qs.get("state") or [None])[0]
        if code or state:
            return (code, state)
    except Exception:
        pass

    if "#" in raw:
        code, state = raw.split("#", 1)
        return (code or None, state or None)

    if "code=" in raw:
        try:
            qs = parse_qs(raw)
            return (qs.get("code", [None])[0], qs.get("state", [None])[0])
        except Exception:
            pass

    return raw, None


def exchange_authorization_code(
    code: str,
    code_verifier: str,
    redirect_uri: str = REDIRECT_URI,
    proxies: Optional[Dict[str, str]] = None,
    timeout: float = 30.0,
) -> CodexOAuthTokens:
    resp = requests.post(
        TOKEN_URL,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={
            "grant_type": "authorization_code",
            "client_id": CLIENT_ID,
            "code": code,
            "code_verifier": code_verifier,
            "redirect_uri": redirect_uri,
        },
        timeout=timeout,
        proxies=proxies,
    )
    resp.raise_for_status()
    data = resp.json() if resp.content else {}
    access = data.get("access_token")
    refresh = data.get("refresh_token")
    expires_in = data.get("expires_in")
    try:
        expires_in_int = int(expires_in)
    except Exception:
        expires_in_int = 0
    if not access or not refresh or expires_in_int <= 0:
        raise RuntimeError("Token exchange response missing required fields")
    return CodexOAuthTokens(
        access=str(access),
        refresh=str(refresh),
        expires_at_ms=int(time.time() * 1000) + (expires_in_int * 1000),
    )


def refresh_access_token(
    refresh_token: str,
    proxies: Optional[Dict[str, str]] = None,
    timeout: float = 30.0,
) -> CodexOAuthTokens:
    resp = requests.post(
        TOKEN_URL,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": CLIENT_ID,
        },
        timeout=timeout,
        proxies=proxies,
    )
    resp.raise_for_status()
    data = resp.json() if resp.content else {}
    access = data.get("access_token")
    refresh = data.get("refresh_token")
    expires_in = data.get("expires_in")
    try:
        expires_in_int = int(expires_in)
    except Exception:
        expires_in_int = 0
    if not access or not refresh or expires_in_int <= 0:
        raise RuntimeError("Token refresh response missing required fields")
    return CodexOAuthTokens(
        access=str(access),
        refresh=str(refresh),
        expires_at_ms=int(time.time() * 1000) + (expires_in_int * 1000),
    )


def decode_jwt_payload(token: str) -> Optional[Dict[str, Any]]:
    parts = (token or "").split(".")
    if len(parts) != 3:
        return None
    payload_b64 = parts[1]
    padding = "=" * (-len(payload_b64) % 4)
    try:
        decoded = base64.urlsafe_b64decode(payload_b64 + padding)
        return json.loads(decoded.decode("utf-8"))
    except Exception:
        return None


def extract_chatgpt_account_id(access_token: str) -> Optional[str]:
    payload = decode_jwt_payload(access_token) or {}
    claim = payload.get(JWT_CLAIM_PATH) or {}
    if isinstance(claim, dict):
        account_id = claim.get(JWT_ACCOUNT_ID_FIELD)
        if account_id:
            return str(account_id)
    return None


class _CallbackServer(ThreadingHTTPServer):
    expected_state: str
    code_event: threading.Event
    code_value: Optional[str]


def start_local_callback_server(state: str, host: str = "127.0.0.1") -> _CallbackServer:
    code_event = threading.Event()

    class Handler(BaseHTTPRequestHandler):
        def log_message(self, format: str, *args: Any) -> None:
            return

        def do_GET(self) -> None:
            parsed = urlparse(self.path)
            if parsed.path != "/auth/callback":
                self.send_response(404)
                self.end_headers()
                self.wfile.write(b"Not Found")
                return

            qs = parse_qs(parsed.query or "")
            received_state = (qs.get("state") or [None])[0]
            code = (qs.get("code") or [None])[0]
            if received_state != self.server.expected_state:
                self.send_response(400)
                self.end_headers()
                self.wfile.write(b"State mismatch")
                return
            if not code:
                self.send_response(400)
                self.end_headers()
                self.wfile.write(b"Missing code")
                return
            self.server.code_value = str(code)
            self.server.code_event.set()
            body = SUCCESS_HTML.encode("utf-8")
            self.send_response(200)
            self.send_header("Content-Type", "text/html; charset=utf-8")
            self.send_header("Content-Length", str(len(body)))
            self.end_headers()
            self.wfile.write(body)

    server = _CallbackServer((host, OAUTH_CALLBACK_PORT), Handler)
    server.expected_state = state
    server.code_event = code_event
    server.code_value = None

    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    return server


def wait_for_callback_code(server: _CallbackServer, timeout_seconds: int = 300) -> Optional[str]:
    if server.code_event.wait(timeout=timeout_seconds):
        return server.code_value
    return None
